/*------------------------------------------------------------------------/
/  Display control module for ST7735
/-------------------------------------------------------------------------/
/
/  Copyright (C) 2014, ChaN, all right reserved.
/
/ * This software is a free software and there is NO WARRANTY.
/ * No restriction on use. You can use, modify and redistribute it for
/   personal, non-profit or commercial products UNDER YOUR RESPONSIBILITY.
/ * Redistributions of source code must retain the above copyright notice.
/
/-------------------------------------------------------------------------*/

#include "disp.h"
#include "LPC15xx.h"

extern void delay_ms (uint32_t);	/* Program delay */

#define LCDC_XOFS	2				/* VRAM offset X */
#define LCDC_YOFS	3				/* VRAM offset Y */
#define	LCDC_CD		GPIO_B40 		/* P1[8] as ST7735 C/D */
#define	LCDC_WR		GPIO_B39 		/* P1[7] as ST7735 WR# */
#define	LCDC_DAT	GPIO_PIN0HH		/* P0[31:24] as ST7735 DATA[7:0] */
#define	DISP_WCMD(d) { LCDC_DAT=(d); LCDC_CD=0; LCDC_WR=0; LCDC_WR=1; LCDC_CD=1; }	/* Write a command byte to the display controller */
#define	DISP_WDAT(d) { LCDC_DAT=(d); LCDC_WR=0; LCDC_WR=1; }	/* Write a data byte to the display controller */
#define	DISP_WPIX(d) { DISP_WDAT((d)>>8); DISP_WDAT(d); }	/* Write a pixel data in RGB565 to the display controller */

#if DISP_USE_DBCS
static uint8_t Sjis;	/* Sjis 1st byte */
#endif


/*-----------------------------------------------------*/
/* Set rectangular area to be transferred              */
/*-----------------------------------------------------*/

void disp_setrect (
	int left,		/* Left end (0..DISP_XS-1) */
	int top,		/* Top end (0..DISP_XS-1) */
	int right,		/* Right end (left..DISP_XS-1) */
	int bottom		/* Bottom end (top..DISP_YS-1) */
)
{
	left += LCDC_XOFS;
	right += LCDC_XOFS;
	top += LCDC_YOFS;
	bottom += LCDC_YOFS;

	DISP_WCMD(0x2A);		/* Set column range */
	DISP_WDAT(left >> 8); DISP_WDAT(left);
	DISP_WDAT(right >> 8); DISP_WDAT(right);

	DISP_WCMD(0x2B);		/* Set row range */
	DISP_WDAT(top >> 8); DISP_WDAT(top);
	DISP_WDAT(bottom >> 8); DISP_WDAT(bottom);

	DISP_WCMD(0x2C);		/* Data write mode */
}



void disp_pixel (
	uint16_t pix,
	int n
)
{
	while (n--) {
		DISP_WPIX(pix);
	}
}



void disp_block_pixel (
	const uint16_t* pix,
	int n
)
{
	uint16_t d;


	do {
		d = *pix++;
		DISP_WPIX(d);
	} while (--n);
}



/*-----------------------------------------------------*/
/* Initialize display module (ST7735 LCDC)             */
/*-----------------------------------------------------*/

void disp_power_on (void)
{
	static const uint8_t st7735_initdata[] = {
		0x01, 0,					/* Software reset (a 50ms delay is needed after this command) */
		0x11, 0,					/* Sleep out (a 120ms delay is needed after this command) */
#if 1 /* ST7735 */
		0xFF, 3, 0x40, 0x03, 0x1A,	/* VCOM4 setting */
		0xD9, 1, 0x60,
		0xC7, 1, 0x90,
		0xB1, 3, 0x04, 0x25, 0x18,	/* Frame rate for normal mode */
		0xB2, 3, 0x04, 0x25, 0x18,	/* Frame rate for idle mode */
		0xB3, 6, 0x04, 0x25, 0x16, 0x04, 0x25, 0x18, 	/* Frame rate for partial mode */
		0xB4, 1, 0x03, 				/* Display inversion control */
		0xB6, 2, 0x15, 0x02,
		0xC0, 2, 0x02, 0x70,		/* Power contorl 1 */
		0xC1, 1, 0x07,				/* Power contorl 2 */
		0xC2, 2, 0x01, 0x01,		/* Power contorl 3 */
		0xC3, 2, 0x02, 0x07,		/* Power control 4 */
		0xC4, 2, 0x02, 0x04,		/* Power contorl 5 */
		0xFC, 2, 0x11, 0x17,		/* Power contorl 6 */
		0xC5, 2, 0x3C, 0x4F,		/* VCOM control 1 */
		0x36, 1, 0xC8,				/* Memory data access contorl */
		0xE0, 16, 0x06, 0x0E, 0x05, 0x20, 0x27, 0x23, 0x1C, 0x21, 0x20, 0x1C, 0x26, 0x2F, 0x00, 0x03, 0x00, 0x24,	/* Gamma settings */
		0xE1, 16, 0x06, 0x10, 0x05, 0x21, 0x27, 0x22, 0x1C, 0x21, 0x1F, 0x1D, 0x27, 0x2F, 0x05, 0x03, 0x00, 0x3F,	/* Gamma settings */
#else /* ST7735S */
		0xB1, 3, 0x02, 0x35, 0x36,	/* Frame rate for normal mode */
		0xB2, 3, 0x02, 0x35, 0x36,	/* Frame rate for idle mode */
		0xB3, 6, 0x02, 0x35, 0x36, 0x02, 0x35, 0x36, 	/* Frame rate for partial mode */
		0xB4, 1, 0x03, 				/* Display inversion control */
		0xC0, 3, 0xA2, 0x02, 0x84,	/* Power contorl 1 */
		0xC1, 1, 0xC5,				/* Power contorl 2 */
		0xC2, 2, 0x0D, 0x00,		/* Power control 3 */
		0xC3, 2, 0x8D, 0xEA,		/* Power control 4 */
		0xC4, 2, 0x8D, 0xEE,		/* Power contorl 5 */
		0xC5, 1, 0x05,				/* VCOM control 1 */
		0x36, 1, 0xC8,				/* Memory data access contorl */
		0xE0, 16, 0x03, 0x1B, 0x09, 0x0E, 0x32, 0x2D, 0x28, 0x2C, 0x2B, 0x29, 0x30, 0x3B, 0x00, 0x01, 0x02, 0x10,	/* Gamma settings */
		0xE1, 16, 0x03, 0x1B, 0x09, 0x0E, 0x32, 0x2E, 0x28, 0x2C, 0x2B, 0x28, 0x31, 0x3C, 0x00, 0x00, 0x02, 0x10,	/* Gamma settings */
#endif
		0x3A, 1, 0x05,				/* Interface pixel format (RGB565) */
		0x29, 0,					/* Display on */
		0
	};
	const uint8_t *p;
	uint8_t cmd, dat, np;
	volatile uint32_t *ioc;
	uint32_t n;


	/* Initialize display module control port */
	GPIO_DIR0HH |= 0xFF;	/* P0[31:24]  |D7|D6|D5|D4|D3|D2|D1|D0| */
	GPIO_DIR1L  |= 0x0180;	/* P1[15:0]   |-|-|-|-|-|-|-|CD|WR#|-|-|-|-|-|-|-| */
	GPIO_DIR2LH |= 0x20;	/* P2[15:8]   |-|-|RES#|-|-|-|-|-| */

	for (ioc = &PIO0_24; ioc <= &PIO0_31; *ioc++ = 0x0080) ;	/* Disable P0[31:24] pull-ups */

	/* Reset LCD module */
	LCDC_WR = 1;
	GPIO_CLR2LH = _BV(5); 		/* LCDC RES# = L */
	delay_ms(1);
	GPIO_SET2LH = _BV(5); 		/* LCDC RES# = H */
	delay_ms(120);

	/* Send initialization data */
	for (p = st7735_initdata; *p; ) {
		cmd = *p++; np = *p++;
		DISP_WCMD(cmd);
		while (np--) {
			dat = *p++;
			DISP_WDAT(dat);
		}
		if (cmd == 0x01) delay_ms(50);
		if (cmd == 0x10) delay_ms(120);
	}

	/* Clear screen and Display ON */
	disp_setrect(0, 0, DISP_XS - 1, DISP_YS - 1);
	for (n = 0; n < DISP_XS * DISP_YS; n++) {
		DISP_WPIX(0);
	}
}



/*-----------------------------------------------------*/
/* Draw a horizontal/vertical line                     */
/*-----------------------------------------------------*/

static
void draw_vhl (
	BMPDEV* bmd,	/* Bitmap device */
	int x,			/* Start position X */
	int y,			/* Start position Y */
	int len,		/* Length */
	int dir			/* 0:right, 1:down */
)
{
	uint8_t *bp, bd, bx, bm, ss, sr;
	uint16_t *wp, wd, js;


	if (x >= bmd->width) return;
	if (x < 0) {
		if (dir) return;
		len += x; x = 0;
	}
	if (y >= bmd->height) return;
	if (y < 0) {
		if (!dir) return;
		len += y; y = 0;
	}
	if (len < 1) return;
	if (dir) {
		if (y + len > bmd->height) len = bmd->height - y;
	} else {
		if (x + len > bmd->width) len = bmd->width - x;
	}

	if (!bmd->bitmap) {	/* Physical display */
		if (dir)
			disp_setrect(x, y, x, y + len - 1);
		else
			disp_setrect(x, y, x + len - 1, y);
		do {
			DISP_WPIX(bmd->color);
		} while (--len);
		return;
	}

	switch (bmd->bpp) {
	case 1:
	case 4:
		if (bmd->bpp == 1) {
			ss = 1; sr = 7; js = bmd->width / 8;
			bp = (uint8_t*)bmd->bitmap + (bmd->width / 8 * y + x / 8);
			bm = 0x80 >> x % 8;
			bx = ((bmd->color & 0x01) << 7) >> x % 8;
		} else {
			ss = 4; sr = 4; js = bmd->width / 2;
			bp = (uint8_t*)bmd->bitmap + (bmd->width / 2 * y + x / 2);
			bm = 0xF0 >> (x % 2) * 4;
			bx = ((bmd->color & 0x0F) << 4) >> (x % 2) * 4;
		}
		do {
			switch (bmd->mode) {
			case 0: bd = (*bp & ~bm) | bx; break;
			case 1: bd = *bp & (~bm | bx); break;
			case 2: bd = *bp | bx; break;
			default: bd = *bp ^ bx;
			}
			*bp = bd;
			if (dir) {
				bp += js;
			} else {
				if (bm & 1) {
					bm <<= sr; bx <<= sr; bp++;
				} else {
					bm >>= ss; bx >>= ss;
				}
			}
		} while (--len);
		break;
	case 8:
		bp = (uint8_t*)bmd->bitmap + (bmd->width * y + x);
		do {
			switch (bmd->mode) {
			case 0: bd = bmd->color; break;
			case 1: bd = *bp & bmd->color; break;
			case 2: bd = *bp | bmd->color; break;
			default: bd = *bp ^ bmd->color;
			}
			*bp = bd;
			bp += dir ? bmd->width : 1;
		} while (--len);
		break;
	case 16:
		wp = (uint16_t*)bmd->bitmap + (bmd->width * y + x);
		do {
			switch (bmd->mode) {
			case 0: wd = bmd->color; break;
			case 1: wd = *wp & bmd->color; break;
			case 2: wd = *wp | bmd->color; break;
			default: wd = *wp ^ bmd->color;
			}
			*wp = wd;
			wp += dir ? bmd->width : 1;
		} while (--len);
	}
}



/*-----------------------------------------------------*/
/* Initialize screen                                   */
/*-----------------------------------------------------*/

int disp_init (
	BMPDEV* bmd,	/* Ptr to the screen descriptor */
	int width,		/* Screen width */
	int height,		/* Screen height */
	uint8_t bpp		/* Color depth (1, 4, 8 or 16) */
)
{
	uint32_t n;
	uint8_t *bp;
	uint16_t *wp;


	if (!bmd->bitmap) {	/* Physical display device */
		disp_power_on();

	} else {			/* On-memory bitmap */
		n = width;
		switch (bpp) {
		case 1:
			n /= 8;
			break;
		case 4:
			n /= 2;
			break;
		case 8: break;
		case 16: break;
		default: n = 0;
		}
		if (n == 0 || height < 0 || width < 0) return 0;
		n *= height;
		if (bpp != 16) {
			bp = (uint8_t*)bmd->bitmap;
			do *bp++ = 0; while (--n);
		} else {
			wp = (uint16_t*)bmd->bitmap;
			do *wp++ = 0; while (--n);
		}
	}

	bmd->width = width;
	bmd->height = height;
	bmd->bpp = bpp;
	bmd->color = 0;
	bmd->x = bmd->y = 0;
	bmd->mode = 0;
	bmd->font_s = 0;
#if DISP_USE_DBCS
	bmd->font_d = 0;
	Sjis = 0;
#endif
	return 1;
}



/*-----------------------------------------------------*/
/* Draw a hollow rectangular                           */
/*-----------------------------------------------------*/

void disp_box (
	BMPDEV* bmd,	/* Ptr to the screen descriptor */
	int x1,			/* X position of an angle */
	int y1,			/* Y position of an angle */
	int x2,			/* X position of an opposite angle */
	int y2			/* Y position of an opposite angle */
)
{
	int n;


	if (x2 < x1) {n = x1; x1 = x2; x2 = n; }
	if (y2 < y1) {n = y1; y1 = y2; y2 = n; }

	draw_vhl(bmd, x1, y1, x2 - x1 + 1, 0);
	if (y1 == y2) return;
	draw_vhl(bmd, x1, y2, x2 - x1 + 1, 0);
	if (y2 < y1 + 2) return;
	draw_vhl(bmd, x1, y1 + 1, y2 - y1 - 1, 1);
	if (x1 == x2) return;
	draw_vhl(bmd, x2, y1 + 1, y2 - y1 - 1, 1);
}



/*-----------------------------------------------------*/
/* Draw a solid rectangular                           */
/*-----------------------------------------------------*/

void disp_fill (
	BMPDEV* bmd,	/* Ptr to the screen descriptor */
	int x1,			/* X position of an angle */
	int y1,			/* Y position of an angle */
	int x2,			/* X position of an opposite angle */
	int y2			/* Y position of an opposite angle */
)
{
	int n;


	if (x2 < x1) {n = x1; x1 = x2; x2 = n; }
	if (y2 < y1) {n = y1; y1 = y2; y2 = n; }

	for (n = y1; n <= y2; n++)
		draw_vhl(bmd, x1, n, x2 - x1 + 1, 0);
}



/*-----------------------------------------------------*/
/* Draw a point                                        */
/*-----------------------------------------------------*/

void disp_point (
	BMPDEV* bmd,	/* Ptr to the screen descriptor */
	int x,			/* Position X */
	int y			/* Position Y */
)
{
	uint8_t *bp, bx, bm, bd;
	uint16_t *wp;


	if ((unsigned int)x >= (unsigned int)bmd->width) return;
	if ((unsigned int)y >= (unsigned int)bmd->height) return;

	if (!bmd->bitmap) {	/* Physical display */
		disp_setrect(x, y, x, y);
		DISP_WPIX(bmd->color);
		return;
	}

	switch (bmd->bpp) {
	case 1:
	case 4:
		if (bmd->bpp == 1) {
			bp = (uint8_t*)bmd->bitmap + (bmd->width / 8 * y + x / 8);
			bm = 0x80 >> x % 8;
			bx = ((bmd->color & 1) << 7) >> x % 8;
		} else {
			bp = (uint8_t*)bmd->bitmap + (bmd->width / 2 * y + x / 2);
			bm = 0xF0 >> (x % 2) * 4;
			bx = ((bmd->color & 15) << 4) >> (x % 2) * 4;
		}
		switch (bmd->mode) {
		case 0: bd = (*bp & ~bm) | bx; break;
		case 1: bd = *bp & (~bm | bx); break;
		case 2: bd = *bp | bx; break;
		default: bd = *bp ^ bx;
		}
		*bp = bd;
		break;
	case 8:
		bp = (uint8_t*)bmd->bitmap + (bmd->width * y + x);
		switch (bmd->mode) {
		case 0: *bp = bmd->color; break;
		case 1: *bp &= bmd->color; break;
		case 2: *bp |= bmd->color; break;
		default: *bp ^= bmd->color;
		}
		break;
	case 16:
		wp = (uint16_t*)bmd->bitmap + (bmd->width * y + x);
		switch (bmd->mode) {
		case 0: *wp = bmd->color; break;
		case 1: *wp &= bmd->color; break;
		case 2: *wp |= bmd->color; break;
		default: *wp ^= bmd->color;
		}
	}
}



/*-----------------------------------------------------*/
/* Draw a line from current position                   */
/*-----------------------------------------------------*/

void disp_lineto (
	BMPDEV* bmd,	/* Ptr to the screen descriptor */
	int x,			/* X position of line to (the last point is not drawn) */
	int y			/* Y position of line to */
)
{
	int32_t xr, yr, xd, yd;
	int ctr, xp, yp, xda, yda;


	xp = bmd->x; bmd->x = x;
	yp = bmd->y; bmd->y = y;
	xd = x - xp; yd = y - yp;
	if (!xd) {
		ctr = yd;
		if (ctr) {
			if (ctr < 0) {
				yp = y + 1; ctr = -ctr;
			}
			draw_vhl(bmd, xp, yp, ctr, 1);
		}
		return;
	}
	if (!yd) {
		ctr = xd;
		if (ctr < 0) {
			xp = x + 1; ctr = -ctr;
		}
		draw_vhl(bmd, xp, yp, ctr, 0);
		return;
	}
	xda = xd < 0 ? -xd : xd;
	yda = yd < 0 ? -yd : yd;
	if (xda >= yda) {
		ctr = xda;
		yd = (yd << 16) / xda;
		xd = (xd < 0 ? -1 : 1) << 16;
	} else {
		ctr = yda;
		xd = (xd << 16) / yda;
		yd = (yd < 0 ? -1 : 1) << 16;
	}
	xr = ((int32_t)xp << 16) + 32768;
	yr = ((int32_t)yp << 16) + 32768;
	do {
		disp_point(bmd, xr >> 16, yr >> 16);
		xr += xd; yr += yd;
	} while (--ctr);
}



/*-----------------------------------------------------*/
/* Transfer bitmap image to the physical display       */
/*-----------------------------------------------------*/

void disp_transfer (
	BMPDEV* s_bmd,	/* Source bitmap */
	int s_left,		/* Top of src rect (0..s_bmd->width-1) */
	int s_top,		/* Left of src rect (0..s_bmd->width-1, >= s_left) */
	int s_right,	/* Right of src rect (0..s_bmd->height-1) */
	int s_bottom,	/* Bottom of src rect (0..s_bmd->height-1, >= s_top) */
	int d_left,		/* Top of dst rect (-32768..32767) */
	int d_top		/* Left of dst rect (-32768..32767) */
)
{
	int yc, xc, hc;
	uint8_t ss;
	uint16_t wd;
	const uint8_t *bp;
	const uint16_t *pal, *wp;


	if (!s_bmd->bitmap) return;
	if (s_left < 0 || s_left > s_right || s_right >= s_bmd->width) return;
	if (s_top < 0 || s_top > s_bottom || s_bottom >= s_bmd->height) return;
	xc = s_right - s_left + 1;
	if (d_left >= DISP_XS || d_left + xc <= 0) return;
	if (d_left < 0) { s_left -= d_left; xc += d_left; d_left = 0; }
	yc = s_bottom - s_top + 1;
	if (d_top >= DISP_YS || d_top + yc <= 0) return;
	if (d_top < 0) { s_top -= d_top; yc += d_top; d_top = 0; }
	if (d_left + xc > DISP_XS) xc = DISP_XS - d_left;
	if (d_top + yc > DISP_YS) yc = DISP_YS - d_top;

	disp_setrect(d_left, d_top, d_left + xc - 1, d_top + yc - 1);	/* Set rectangular area to be transferred */
	pal = s_bmd->palette;
	switch (s_bmd->bpp) {
	case 1:
		do {
			bp = (uint8_t*)s_bmd->bitmap + (s_bmd->width / 8 * s_top + s_left / 8);
			ss = 7 - (s_left % 8);
			hc = xc;
			do {
				wd = pal[*bp >> ss & 0x01];
				DISP_WPIX(wd);
				if (!ss) { 
					bp++; ss = 8;
				}
				ss--;
			} while (--hc);
			s_top++;
		} while (--yc);
		break;
	case 4:
		do {
			bp = (uint8_t*)s_bmd->bitmap + (s_bmd->width / 2 * s_top + s_left / 2);
			ss = (1 - (s_left % 2)) * 4;
			hc = xc;
			do {
				wd = pal[*bp >> ss & 0x0F];
				DISP_WPIX(wd);
				if (ss ^= 4) bp++;
			} while (--hc);
			s_top++;
		} while (--yc);
		break;
	case 8:
		do {
			bp = (uint8_t*)s_bmd->bitmap + (s_bmd->width * s_top + s_left);
			hc = xc;
			do {
				wd = pal[*bp++];
				DISP_WPIX(wd);
			} while (--hc);
			s_top++;
		} while (--yc);
		break;
	case 16:
		do {
			wp = (uint16_t*)s_bmd->bitmap + (s_bmd->width * s_top + s_left);
			hc = xc;
			do {
				wd = *wp++;
				DISP_WPIX(wd);
			} while (--hc);
			s_top++;
		} while (--yc);
	}
}



/*-----------------------------------------------------*/
/* Put a text character                                */
/*-----------------------------------------------------*/

void disp_putc (
	BMPDEV* bmd,
	uint8_t chr		/* Character to be output (kanji chars are given in two disp_putc sequence) */
)
{
	const uint8_t *fnt;
	uint8_t b, d;
	uint16_t dchr;
	int h, wc, w, wb, x, y, fofs;


	fnt = bmd->font_s;
	if (!fnt) return;	/* Exit if no font registerd */

	if (chr < 0x20) {	/* Processes the control character */
#if DISP_USE_DBCS
		Sjis = 0;
#endif
		switch (chr) {
		case '\n':		/* Next row */
			bmd->y += fnt[15];
			/* follow next case */
		case '\r':		/* Origin column */
			bmd->x = 0;
			return;
		case '\b':		/* Reverse full column */
			bmd->x -= fnt[14];
			if (bmd->x < 0) bmd->x = 0;
			return;
		case '\x10':	/* Reverse quater column */
			bmd->x -= fnt[14] / 4;
			return;
		case '\x11':	/* Reverse half column */
			bmd->x += fnt[14] / 2;
			return;
		case '\x12':	/* Forward quater column */
			bmd->x += fnt[14] / 4;
			return;
		case '\x13':	/* Forward half column */
			bmd->x += fnt[14] / 2;
			return;
		}
	}

	/* Exit if current position is out of bitmap */
	if ((unsigned int)bmd->x >= (unsigned int)bmd->width || (unsigned int)bmd->y >= (unsigned int)bmd->height) return;

#if DISP_USE_DBCS
	if (Sjis) {	/* This is sjis trailing byte */
		uint16_t bchr, rs, re;
		int ri, i;

		dchr = Sjis * 256 + chr; 
		Sjis = 0;
		fnt = bmd->font_d;	/* Switch to double byte font */
		i = fnt[17];	/* Number of code blocks */
		ri = 18;		/* Start of code block table */
		bchr = 0;		/* Number of chars in previous blocks */
		while (i) {		/* Find the code in the code blocks */
			rs = fnt[ri + 0] + fnt[ri + 1] * 256;	/* Start of a block */
			re = fnt[ri + 2] + fnt[ri + 3] * 256;	/* End of a block */
			if (dchr >= rs && dchr <= re) break;	/* Is the code in the block? */
			bchr += re - rs + 1; ri += 4; i--;		/* Next block */
		}
		if (!i) {	/* Code not found */
			bmd->x += fnt[14];		/* Put a transparent character */
			return;
		}
		dchr = dchr - rs + bchr;	/* Character offset in the font area */
		fofs = 18 + fnt[17] * 4;	/* Font area start address */
	} else {
		/* Check if this is sjis leading byte */
		if (bmd->font_d && (((uint8_t)(chr - 0x81) <= 0x1E) || ((uint8_t)(chr - 0xE0) <= 0x1C))) {
			Sjis = chr;	/* Next is sjis trailing byte */
			return;
		}
		dchr = chr;
		fofs = 17;		/* Font area start address */
	}
#else
	dchr = chr;
	fofs = 17;
#endif



	h = fnt[15]; w = fnt[14]; wb = (w + 7) / 8;	/* Font size: height, dot width and byte width */
	fnt += fofs + dchr * wb * h;				/* Goto start of the bitmap */

	d = 0; y = bmd->y;
	do {
		x = bmd->x;
		wc = w; b = 1;
		do {
			if (--b == 0) { b = 8; d = *fnt++; }
			if (d & 0x80) disp_point(bmd, x, y);
			d <<= 1; x++;
		} while (--wc);
		y++;		/* Next raster */
	} while (--h);

	bmd->x += w;	/* Update current position */
}


