

#include "LPC15xx.h"
#include "uart15xx.h"
#include "xprintf.h"
#include "disp.h"
#include "intfft.h"


/* Timings */
#define F_SYSCLK	18000000	/* System clock frequency */
#define F_SAMPLE	16000		/* Sampling frequency */
#define N_WFM		128			/* Number of re-samples of a cycle */
#define N_ALT		64			/* Number of input alts time measured for frequency counting */
#define N_AVE		8			/* Number of cycle averagings for a waveform capture */

/* Waveform window */
#define WFM_XOFS	0
#define WFM_YOFS	32
#define WFM_HEIGHT	96

/* Current roll window */
#define ROLL_XOFS	4
#define ROLL_YOFS	40
#define ROLL_WIDTH	96
#define ROLL_HEIGHT	60

#define V_CAL		30			/* Calibration voltage */
#define I_CAL		2			/* Calibration current */
#define STRING(x)	#x

/* Bit definitions of BtnDown and BtnStat */
#define BTN_MODE	_BV(0)
#define BTN_HOLD	_BV(1)
#define BTN_UP		_BV(3)
#define BTN_DOWN	_BV(4)

/* Flags bit definitions */
#define FLAG_SAMPLE	_BV(0)
#define FLAG_CYCLE	_BV(1)
#define FLAG_FREQ	_BV(2)
#define FLAG_OVF	_BV(4)
#define FLAG_RUN	_BV(7)



/* Palette tables for 4/1 bpp bitmap */
const uint16_t Pal16[16] = {RGB16(0,0,0), RGB16(0,0,64), RGB16(0,64,0), RGB16(0,64,64), RGB16(64,0,0), RGB16(64,0,64), RGB16(64,64,0), RGB16(64,64,64), RGB16(80,80,80), RGB16(0,0,255), RGB16(0,255,0), RGB16(0,255,255), RGB16(255,0,0), RGB16(255,0,255), RGB16(255,255,0), RGB16(255,255,255)};
const uint16_t Pal2bw[2] = {RGB16(0,0,0), RGB16(255,255,255)};
const uint16_t Pal2bl[2] = {RGB16(0,0,80), RGB16(255,255,255)};
const uint16_t Pal2bm[2] = {RGB16(0,0,80), RGB16(255,64,0)};

/* Import FONTX2 files as byte array */
IMPORT_BIN(".rodata", "rc/fnt12x24.fnt", FontHan24);	/* const uint8_t FontHan24[] */
IMPORT_BIN(".rodata", "rc/fnt8x16.fnt", FontHan16);		/* const uint8_t FontHan16[] */
IMPORT_BIN(".rodata", "rc/fnt6x8.fnt", FontHan8);		/* const uint8_t FontHan8[] */
extern const uint8_t FontHan24[], FontHan16[], FontHan8[];

/* Import image files as byte array */
IMPORT_BIN(".rodata", "rc/heco.bin", ImgLogo);		/* const uint8_t ImgLogo[] (128x32/16bpp) */
IMPORT_BIN(".rodata", "rc/heco2.bin", ImgLogo2);	/* const uint8_t ImgLogo2[] (128x32/16bpp) */
extern const uint8_t ImgLogo[], ImgLogo2[];



volatile uint32_t Timer;			/* F_SAMPLE Hz increment timer */

volatile uint8_t Flags;				/* b0:New sample, b1:New cycle, b2:Frequency updated, b4:Input over range, b7:Run */
volatile uint8_t Capture;			/* Capture command: b7:Request to start, b6-b0:Cycles to capture */
volatile uint8_t BtnDown, BtnStat;	/* Button status flags */

volatile int32_t Adc_V, Adc_I;		/* Sampling data (calibrated) 32768/4800V, 32768/30A */
volatile int64_t AccJoule;			/* Accumulated energy (real-time) */
volatile uint32_t TimeRT;			/* Duration from start (real-time) */
volatile uint16_t FreqRT;			/* Measured frequency * 100Hz */

/* Cycle current fifo for current profile mode */
volatile int64_t CycleCurrent[8];	/* Cycle current (to be sqrted and get 32768/30A) */
volatile uint8_t NumCC, RpCC, WpCC;

int32_t Cal[5];		/* Analog path calibration data (vofs, vgain, iofs, igain, sum) */

/* Waveform and statistics updated by capture_wfm() */
int32_t Waveform[N_WFM][2];			/* Waveform capture buffer {voltage, current} */
int32_t Volt_pk, Volt_rms;			/* Voltage: Peak, RMS */
int32_t Curr_pk, Curr_rms;			/* Current: Peak, RMS */
int32_t Pow_eff, Pow_app, Pow_fac;	/* Effective power, Apparent power, Power factor */
uint16_t Freq;						/* Frequency at end of capture */
uint32_t Time;						/* Duration at end of capture */
int64_t Joule;						/* WH value at end of capture */



BMPDEV BmdLcd;				/* Bitmap object of physical display */
BMPDEV BmdMem, *PutchBmd;	/* Bitmap object of on-memory screen, destination of putch_bmd() */
uint8_t Bitmap[8192] __attribute__ ((aligned (4)));		/* On memory bitmap */





void delay_ms (uint32_t ms)
{
	Timer = 0;
	while (Timer < ms * (F_SAMPLE / 1000)) ;
}



void putch_bmd (uint8_t chr)
{
	if (chr == '.' || chr == ':') disp_putc(PutchBmd, '\x10');
	disp_putc(PutchBmd, chr);
	if (chr == '.') disp_putc(PutchBmd, '\x10');
}



uint8_t get_btn (void)
{
	uint8_t d;

	__disable_irq();
	d = BtnDown;
	BtnDown = 0;
	__enable_irq();
	return d;
}



int32_t get_adc (volatile int32_t* p, int32_t ave)
{
	int32_t d, n;


	for (d = n = 0; n < ave; n++) {
		__disable_irq();
		Flags &= ~FLAG_SAMPLE;
		__enable_irq();
		while (!(Flags & FLAG_SAMPLE)) ;
		d += *p;
	}
	return d / ave;
}



/*--------------------------------------------------------------------/
/ Hardware Setup                                                      /
/--------------------------------------------------------------------*/

static
void ioinit (void)
{
	uint32_t t;
	uint32_t cmd[5], res[5], *p;


	/* Set power profile */
	p = (uint32_t*)*(uint32_t*)0x3000200;	/* Get API table */
	p = (uint32_t*)p[3];					/* Get power API */
	cmd[0] = 36; cmd[1] = 3; cmd[2] = 18; cmd[3] = 0;
	((void (*)(uint32_t[], uint32_t[]))p[1])(cmd, res);	/* Set power profile (low-current) */

	/* Disable SWD function to assign button input */
	__disable_pin_function(SWCLK_EN | SWDIO_EN);

	/* Enable clock to the integrated peripherals to be used */
	__enable_ahbclk(PCEEPROM | PCSWM | PCIOCON | PCGPIO0 | PCGPIO1 | PCGPIO2);


	/* Enable USART0 module */
//	__attach_function(UART0_RXD_I, 0, 13);	/* UART0_RXD -> P0_13 */
//	__attach_function(UART0_TXD_O, 0, 18);	/* UART0_TXD -> P0_18 */
//	GPIO_DIR0HL |= _BV(2);					/* P0_18: output */
//	uart0_init(0);


	/* Configure SCT0 as F_SAMPLE Hz square wave generator, interrupt on falling edge */
	__enable_pin_function(SCT0_OUT3_EN);	/* Attach SCT0 OUT3 to PIO0_0 */
	GPIO_DIR0LL |= _BV(0);					/* P0_0: Output (AD7685.CNV) */

	__enable_ahbclk(PCSCT0);			/* Enable SCT0 clock */
	SCT0_CONFIG = 0x00000;				/* Separated mode, clock from sysclk, no auto-limit */
	t = F_SYSCLK / F_SAMPLE - 1;		/* SCT cycle time T */
	SCT0_MATCH0_L = t; SCT0_MATCHREL0_L = t;			/* Match0L at T */
	SCT0_MATCH1_L = t / 2; SCT0_MATCHREL1_L = t / 2;	/* Match1L at T / 2 */
	SCT0_EV0_CTRL = 0x01000;			/* Event0: on match0L */
	SCT0_EV0_STATE = _BV(0);			/* Event0: enabled at state 0 */
	SCT0_EV1_CTRL = 0x01001;			/* Event1: on match1L */
	SCT0_EV1_STATE = _BV(0);			/* Event1: enabled at state 0 */
	SCT0_LIMIT_L = _BV(0);				/* CounterL cleared on Event0 */
	SCT0_OUT3_SET = _BV(0); SCT0_OUT3_CLR = _BV(1);	/* OUT3: set on Event0, cleared on Event1 */
	SCT0_EVEN = _BV(1);					/* Interrupt: on Event1 */
	SCT0_CTRL_L = 0x0008;				/* Start SCT0_L */
	__enable_irqn(SCT0_IRQn);			/* Enable SCT0 interrupt */

	/* Enable SPI0 */
	__attach_function(SPI0_MISO_IO, 0, 1);	/* MISO -> P0_1 (AD7685.SDO) */
	__attach_function(SPI0_SCK_IO, 1, 0);	/* SCK -> P1_0 (AD7685.SCK) */
	GPIO_DIR1LL |= _BV(0);					/* P1_0: output */

	__enable_ahbclk(PCSPI0);	/* Enable SPI0 clock */
	SPI0_DIV = 2 - 1;			/* Fsck = sysclk / 2 */
	SPI0_CFG = 0x005;			/* Enable SPI0 as Master, Mode 0 */


	/* Enable ADC0 */
	__enable_pin_function(ADC0_1_EN);	/* ADC0_1 -> P0_8 */
	PIO0_8 = 0x0080;					/* Disable pull-up on P0_8 pin */
	__attach_function(ADC0_PIN_TRIG0_I, 0, 0);	/* ADC0_PIN_TRIG0 -> P0_0 */

	__enable_ahbclk(PCADC0);	/* Enable ADC0 clock */
	__enable_power(PDADC0);		/* Enable ADC0 power */
	delay_ms(500);								/* Delay 500ms for Vcc stabled */
	ADC0_CTRL = _BV(30) | (F_SYSCLK / 500000);	/* Launch self-calibration process */
	while (ADC0_CTRL & _BV(30)) ;				/* Wait for end of calibration */
	ADC0_CTRL = 0;
	ADC0_SEQA_CTRL = 0x80040002;	/* Select ADC0_1 only */


	__disable_ahbclk(PCSWM);			/* Disable clock for SWM (no longer needed) */
}



/*--------------------------------------------------------------------/
/ Calibration                                                         /
/--------------------------------------------------------------------*/

static
void load_calib (void)
{
	uint32_t cmd[5], res[5];	/* IAP command and parameter */
	uint32_t i;
	int32_t ofs, gain, d;


	BmdLcd.font_s = FontHan16;
	BmdLcd.color = RGB16(255,255,255);
	PutchBmd = &BmdLcd;
	xputs("Hecowatt v1.0\n");

	Flags = FLAG_RUN;

	if (!GPIO_B13) {	/* Is P0_13 low? */
		Cal[0] = 1; Cal[1] = 1024; Cal[2] = 1; Cal[3] = 1024;

		xputs("Set Vin 0V");						/* Voltage path offset adjustment */
		while (get_btn() != BTN_DOWN) ;
		ofs = get_adc(&Adc_V, 100);
		xputs("\nSet Vin " STRING(V_CAL) "V DC");	/* Current path gain adjustment */
		while (get_btn() != BTN_DOWN) ;
		gain = (get_adc(&Adc_V, 100) - ofs) * 1024 / (int32_t)(2048.0 / 300 * V_CAL);
		Cal[0] = ofs; Cal[1] = gain;

		xputs("\nSet Iin 0A");						/* Current path offset adjustment */
		while (get_btn() != BTN_DOWN) ;
		ofs = get_adc(&Adc_I, 100);
		xputs("\nSet Iin" STRING(I_CAL) "A DC");	/* Current path gain adjustment */
		while (get_btn() != BTN_DOWN) ;
		gain = (get_adc(&Adc_I, 100) - ofs) * 1024 / (int32_t)(32768.0 / 30 * I_CAL);
		Cal[2] = ofs; Cal[3] = gain;

		for (i = d = 0; i < sizeof Cal / sizeof Cal[0] - 1; d += Cal[i++]) ;	/* Create check sum */
		Cal[i] = d + 5555;

		/* Save calibration data to the EEPROM */
		cmd[0] = IAP_EEPROM_WRITE;	/* EEPROM write command */
		cmd[1] = (uint32_t)0x40;	/* Destination: Top of EEPROM user area */
		cmd[2] = (uint32_t)Cal;		/* Source: Calibration data */
		cmd[3] = sizeof Cal;		/* Number of bytes to write */
		cmd[4] = 12000000;			/* System clock to EEPROM (dummy) */
		IAP_ENTRY(cmd, res);
		if (res[0] == 0) xputs("\nData saved.\n");
		for (;;) ;
	}

	/* Load calibration data from the EEPROM */
	cmd[0] = IAP_EEPROM_READ;	/* EEPROM read command */
	cmd[1] = (uint32_t)0x40;	/* Source: Top of EEPROM user area */
	cmd[2] = (uint32_t)Cal;		/* Destination: Calibration data array */
	cmd[3] = sizeof Cal;		/* Number of bytes to read */
	cmd[4] = 12000000;			/* System clock to EEPROM (dummy) */
	IAP_ENTRY(cmd, res);
	for (i = d = 0; i < sizeof Cal / sizeof Cal[0] - 1; d += Cal[i++]) ;	/* Check sum */
	if (d + 5555 != Cal[i]) {
		xputs("\nCalibration data\nis collapsed!\n");
		for (;;) ;
	}

	xputs("\n\nWait a Minute...");
	for (i = 0; i < 2; i++) {
		__disable_irq();
		Flags &= ~(FLAG_FREQ | FLAG_OVF);
		__enable_irq();
		while (!(Flags & FLAG_FREQ)) ;
	}

	AccJoule = 0;
	TimeRT = 0;
}




/*--------------------------------------------------------------*/
/* Create a numeral string to be displayed from given A-D value */
/*--------------------------------------------------------------*/

char* mkval (
	int64_t val,
	int unit	/* 0:V, 1:A, 2:W/VA, 3:kWh, 4:Hz */
)
{
	static char str[16];
	int32_t v;


	if (val < 0) val = 0;

	switch (unit) {
	case 0:	/* Voltage [V] */
		v = (int32_t)val * 3000 / 2048;	/* Voltage in unit of 0.1V */
		xsprintf(str, "\x13%3d.%d", v / 10, v % 10);	/* 000.0 */
		break;
	case 1:	/* Current [A] */
		v = (int32_t)val * 30000 / 32768;	/* Current in unit of 1mA */
		if (v < 10000) {
			xsprintf(str, "\x13%d.%03d", v / 1000, v % 1000);	/* 0.000 */
		} else {
			v /= 10;
			xsprintf(str, "\x13%2d.%02d", v / 100, v % 100);	/* 00.00 */
		}
		break;
	case 2:	/* Power [W/VA] */
		v = val * 5625 / 4194304;	/* Wattage in unit of 0.1W */
		if (v < 10000)
			xsprintf(str, "\x13%3d.%d", v / 10, v % 10);	/* 000.0 */
		else
			xsprintf(str, "%5d", (v / 5 + 1) / 2);	/* 0000 */
		break;
	case 3:	/* Energy */
		v = val / ((int64_t)F_SAMPLE * 26835456);
		if (v < 1000) {
			xsprintf(str, "\x13.%03d", v);	/* .000 */
		} else {
			if (v < 10000) {
				v = (v / 5 + 1) / 2;
				xsprintf(str, "\x13%d.%02d", v / 100, v % 100);	/* 0.00 */
			} else {
				v = (v / 50 + 1) / 2;
				if (v < 1000) {
					xsprintf(str, "\x13%2d.%d", v / 10, v % 10);	/* 00.0 */
				} else {
					xsprintf(str, "%4d", v);		/* 0000 */
				}
			}
		}
		break;
	case 4:	/* Frequency [Hz] */
		v = val;
		xsprintf(str, "\x13%2d.%02d", v / 100, v % 100);	/* 00.00 */
		break;
	default:
		*str = 0;
	}

	return str;
}



/*------------------------------------------------------------*/
/* Capture a waveform with N_AVE averaging and get statistics */
/*------------------------------------------------------------*/

void capture_wfm (
	uint8_t cmd	/* bit0:Start to capture, bit1:End process */
)
{
	static int valid;
	int32_t idx, v, i, vpk, ipk;
	int64_t vrms, irms, pow;


	if (cmd & _BV(0)) {	/* Start to capture a new waveform */
		Capture = 0;
		for (i = 0; i < N_WFM; i++) {
			Waveform[i][0] = 0;
			Waveform[i][1] = 0;
		}
		Capture = 0x80 | N_AVE;
		valid = 0;
	}

	if (cmd & _BV(1) && !valid) {	/* Get statistics of the waveform */
		valid = 1;
		while (Capture) ;	/* Wait for end of capture */

		vpk = ipk = 0;
		vrms = irms = pow = 0;
		for (idx = 0; idx < N_WFM; idx++) {
			Waveform[idx][0] = v = Waveform[idx][0] / N_AVE;
			Waveform[idx][1] = i = Waveform[idx][1] / N_AVE;
			if (v > vpk) vpk = v;
			if (-v > vpk) vpk = -v;
			if (i > ipk) ipk = i;
			if (-i > ipk) ipk = -i;
			vrms += (int64_t)v * v;
			irms += (int64_t)i * i;
			pow += (int64_t)i * v;
		}

		Volt_pk = vpk; Curr_pk = ipk;
		Pow_eff = pow / N_WFM;
		Volt_rms = isqrt64(vrms / N_WFM);
		Curr_rms = isqrt64(irms / N_WFM);
		Pow_app = Volt_rms * Curr_rms;
		Pow_fac = (Pow_app > 0 && Pow_eff > 0) ? ((int64_t)Pow_eff * 100 + Pow_eff / 2) / Pow_app : 0;
		Joule = AccJoule;
		Freq = FreqRT;
		Time = TimeRT;
	}
}



/*---------------------------------------------------*/
/* Display full statistics captured by capture_wfm() */
/*---------------------------------------------------*/

void rfsh_full (
	int	label	/* 0:Draw values, 1:Draw labels */
)
{
	BMPDEV *bmd = &BmdMem;
	const uint32_t *ps;
	uint32_t n, *pd;

	static const uint8_t lstcolor[6] = {2, 1, 1, 1, 4, 4};


	disp_init(bmd, 128, 32, 16);	/* Create virtual screen for title logo */
	ps = (const uint32_t*)ImgLogo; pd = (uint32_t*)Bitmap;
	for (n = 0; n < 128 * 32 / 2; n++) *pd++ = *ps++;

	if (label) {
		disp_transfer(bmd, 0, 0, 47, 31, 0, 0);

		disp_init(bmd, 64, 96, 4);	/* Create virtual screen for labels */
		for (n = 0; n < 6; n++) {
			bmd->color = lstcolor[n];
			disp_fill(bmd, 0, n *16, 63, n * 16 + 15);
		}
		bmd->font_s = FontHan8;
		bmd->color = 15;
		bmd->x = 4;  bmd->y = 4;  xputs("Current");
		bmd->x = 4;  bmd->y = 16; xputs("Effective");
		bmd->x = 28; bmd->y = 24; xputs("power");
		bmd->x = 4;  bmd->y = 32; xputs("Apparent");
		bmd->x = 28; bmd->y = 40; xputs("power");
		bmd->x = 4;  bmd->y = 48; xputs("Power");
		bmd->x = 22; bmd->y = 56; xputs("factor");
		bmd->x = 4;  bmd->y = 68; xputs("Voltage");
		bmd->x = 4;  bmd->y = 84; xputs("Frequency");
		bmd->palette = Pal16;
		disp_transfer(bmd, 0, 0, 63, 95, 0, 32);
		return;
	}

	bmd->font_s = FontHan24;
	bmd->x = 52; bmd->y = 8; bmd->color = RGB16(255,255,255);
	xprintf("%s", mkval(Joule, 3));
	bmd->y += 8;
	bmd->font_s = FontHan16;
	xputs("\x12" "kWh");
	bmd->font_s = FontHan8;
	bmd->x = 66; bmd->y = 0;
	xprintf("%4u:%02u:%02u", Time / 3600, (Time / 60) % 60, Time % 60);
	disp_transfer(bmd, 48, 0, 127, 31, 48, 0);

	disp_init(bmd, 64, 96, 4);	/* Create virtual screen for each value */
	for (n = 0; n < 6; n++) {
		bmd->color = lstcolor[n];
		disp_fill(bmd, 0, n *16, 63, n * 16 + 15);
	}
	bmd->font_s = FontHan16;
	bmd->color = 15;
	xprintf("%s\x13" "A\n", mkval(Curr_rms, 1));
	xprintf("%s\x13" "W\n", mkval(Pow_eff, 2));
	xprintf("%s\x13" "VA\n", mkval(Pow_app, 2));
	xprintf("%5d\x13" "%%\n", Pow_fac);
	xprintf("%s\x13" "V\n", mkval(Volt_rms, 0));
	xprintf("%s\x13" "Hz\n", mkval(Freq, 4));

	bmd->palette = Pal16;
	disp_transfer(bmd, 0, 0, 63, 95, 64, 32);
}



/*-----------------------------------------------------*/
/* Display simple statistics captured by capture_wfm() */
/*-----------------------------------------------------*/

void rfsh_simple (void)
{
	BMPDEV *bmd = &BmdMem;
	const uint32_t *ps;
	uint32_t n, *pd;


	disp_init(bmd, 128, 32, 16);	/* Create virtual screen for title logo */
	ps = (const uint32_t*)ImgLogo2; pd = (uint32_t*)Bitmap;
	for (n = 0; n < 128 * 32 / 2; n++) *pd++ = *ps++;

	bmd->color = RGB16(255,255,255);
	bmd->font_s = FontHan16;

	bmd->x = 0; bmd->y = 0;
	xprintf("%s\x13" "V", mkval(Volt_rms, 0));
	bmd->x = 0; bmd->y = 16;
	xprintf("%s\x13" "A", mkval(Curr_rms, 1));
	bmd->x = 72; bmd->y = 0;
	xprintf("%s\x13" "W", mkval(Pow_eff, 2));
	bmd->x = 72; bmd->y = 16;
	xprintf("%5d\x13" "%%", Pow_fac);

	disp_transfer(bmd, 0, 0, 127, 31, 0, 0);
}



/*-------------------------*/
/* Display current profile */
/*-------------------------*/

void rfsh_roll (
	uint8_t cmd,	/* bit0:Draw profile window, bit1:Push new data, bit2:Clear roll buffer, bit3:Draw status */
	int mag			/* Display magnify (0..7) */
)
{
	BMPDEV *bmd = &BmdMem;
	static const char *scale[8][6] = {
		{".10", ".08", ".06", ".04", ".02", "0"},
		{".20", ".16", ".12", ".08", ".04", "0"},
		{"0.5", "0.4", "0.3", "0.2", "0.1", "0"},
		{"1.0", "0.8", "0.6", "0.4", "0.2", "0"},
		{"2.0", "1.6", "1.2", "0.8", "0.4", "0"},
		{"5.0", "4.0", "3.0", "2.0", "1.0", "0"},
		{"10", "8", "6", "4", "2", "0"},
		{"20", "16", "12", "8", "4", "0"}
	};
	static const int32_t magtbl[8] = {109, 218, 546, 1092, 2185, 5461, 10922, 21845};
	static int32_t roll_buf[ROLL_WIDTH], ipk;
	static uint16_t ofs, ipk_upd;
	uint32_t i;
	int32_t val;
	int16_t col;



	if (cmd & _BV(1)) {	/* Push a new data into roll buffer */
		while (!NumCC) ;
		i = RpCC;
		val = (int32_t)isqrt64(CycleCurrent[i]);
		roll_buf[ofs] = val;
		ofs = (ofs + 1) % ROLL_WIDTH;
		RpCC = ++i % 8;
		__disable_irq();
		NumCC--;
		__enable_irq();
		if (val > ipk) {
			ipk = val;
			ipk_upd = 1;
		}
	}

	if (cmd & _BV(2)) {	/* Clear roll buffer */
		__disable_irq();
		NumCC = 0; RpCC = 0; WpCC = 0;
		Flags &= ~FLAG_OVF;
		__enable_irq();
		for (i = 0; i < ROLL_WIDTH; roll_buf[i++] = -1) ;
		ipk = -1;
	}

	if (cmd & _BV(3)) {	/* Draw background and scale */
		disp_init(bmd, 32, 96, 1);	/* Create virtual screen */
		bmd->font_s = FontHan8; bmd->color = 1;
		for (i = 0; i < 6; i++) {
			bmd->y += 4;
			xprintf("`%sA\n", scale[mag][i]);
		}
		bmd->palette = Pal2bl;
		disp_transfer(bmd, 0, 0, 27, 95, 100, 32);
		disp_init(bmd, 128, 128, 1);	/* Create virtual screen */
		disp_transfer(bmd, 0, 0, 99, 7, 0, 32);
		disp_transfer(bmd, 0, 0, 3, 71, 0, 32);
		disp_transfer(bmd, 0, 0, 99, 31, 0, 100);
		ipk_upd = 1;
	}

	if (cmd & _BV(0)) {	/* Dislpay current profile */
		for (i = 0; i < ROLL_WIDTH; i++) {
			disp_setrect(ROLL_XOFS + i, ROLL_YOFS, ROLL_XOFS + i, ROLL_YOFS + ROLL_HEIGHT - 1);
			val = roll_buf[(ofs + i) % ROLL_WIDTH];
			if (val < 0) {
				val = ROLL_HEIGHT;
				col = RGB16(0,0,0);
			} else {
				val = val * ROLL_HEIGHT / magtbl[mag];
				if (val >= ROLL_HEIGHT) {
					val = ROLL_HEIGHT;
					col = RGB16(255,0,0);
				} else {
					disp_pixel(RGB16(0,0,0), ROLL_HEIGHT - 1 - val);
					col = RGB16(0,255,0);
					val++;
				}
			}
			disp_pixel(col, val);
		}
	}

	if (ipk_upd && ipk >= 0 && NumCC < 4) {	/* Update peak current */
		disp_init(bmd, 80, 8, 1);	/* Create virtual screen */
		bmd->font_s = FontHan8;
		bmd->color = 1;
		xprintf("Peak:%sA", mkval(ipk, 1));
		bmd->palette = (Flags & FLAG_OVF) ? Pal2bm : Pal2bl;
		disp_transfer(bmd, 0, 0, 79, 7, 4, 106);
		ipk_upd = 0;
	}
}



/*--------------------------------------------*/
/* Display waveform captured by capture_wfm() */
/*--------------------------------------------*/

void rfsh_scope (void)
{
	BMPDEV *bmd = &BmdMem;
	int i, y, vp, ip;


	disp_init(bmd, N_WFM, WFM_HEIGHT, 4);	/* Create virtual screen for scope window */

	vp = Volt_pk;
	if (vp < WFM_HEIGHT / 2) vp = WFM_HEIGHT / 2;
	ip = Curr_pk;
	if (ip < WFM_HEIGHT / 2) ip = WFM_HEIGHT / 2;

	bmd->color = 12;	/* Color: Green-replace */
	for (i = 0; i < N_WFM; i++) {
		y = Waveform[i][0] * -(WFM_HEIGHT / 2) / vp + (WFM_HEIGHT / 2);
		if (i) {
			disp_lineto(bmd, i, y);
		} else {
			bmd->x = i; bmd->y = y;
		}
	}
	disp_point(bmd, i, y);

	bmd->color = 10; bmd->mode = 2;	/* Color: Red-or */
	for (i = 0; i < N_WFM; i++) {
		y = Waveform[i][1] * -(WFM_HEIGHT / 2) / ip + (WFM_HEIGHT / 2);
		if (i) {
			disp_lineto(bmd, i, y);
		} else {
			bmd->x = i; bmd->y = y;
		}
	}
	disp_point(bmd, i, y);

	bmd->color = 8; bmd->mode = 0;	/* Color: Gray-replace */
	for (i = 0; i < N_WFM; i += 5)
		disp_point(bmd, i, WFM_HEIGHT / 2);

	bmd->color = 15;	/* Color: White-replace */
	bmd->x = 0; bmd->y = WFM_HEIGHT - 18;
	bmd->font_s = FontHan8;
	xprintf("Vp:\x10%s\x13V\n", mkval(Volt_pk, 0));
	bmd->y += 2;
	xprintf("Ip:\x10%s\x13" "A\n", mkval(Curr_pk, 1));

	bmd->palette = Pal16;
	disp_transfer(bmd, 0, 0, N_WFM - 1, WFM_HEIGHT - 1, WFM_XOFS, WFM_YOFS);
}



/*----------------------------------------------------*/
/* Display harmonic current captured by capture_wfm() */
/*----------------------------------------------------*/

void rfsh_harmonics (
	int ofs		/* Display row offset (0-11) */
)
{
	BMPDEV *bmd = &BmdMem;
	int16comp_t fftbuf[N_WFM];
	uint16_t fftout[N_WFM];
	int32_t d;
	int i, y;


	/* Apply DFT to the waveform */
	for (i = 0; i < N_WFM; i++) {	/* Load waveform to the FFT buffer */
		d = Waveform[i][1];
		if (d > 32767) d = 32767;
		if (d < -32768) d = -32768;
		fftbuf[i].r = d;
		fftbuf[i].i = 0;
	}
	int16fft_exec(fftbuf);			/* Execute FFT */
	int16fft_output_scalar(fftbuf, fftout);	/* Re-order and convert it to scalar values */

	disp_init(bmd, 128, 96, 1);	/* Create virtual screen for harmonics window */
	bmd->color = 1;
	bmd->font_s = FontHan8;

	for (y = 0; y < 96 / 10; y++) {
		bmd->y = y * 10 + 1;
		if (y + ofs == 0) {		/* DC and Fundamental component */
			d = fftout[N_WFM / 2];
			bmd->x = 2;
			xprintf("DC:%s\x13" "A", mkval(d * 724 / 1024, 1));
			d = fftout[N_WFM / 2 + 1] + fftout[N_WFM / 2 - 1];
			bmd->x = 70;
			xprintf("Fu:%s\x13" "A", mkval(d * 724 / 1024, 1));
		} else {				/* 2nd component and followings... */
			d = fftout[N_WFM / 2 + (y + ofs) * 2] + fftout[N_WFM / 2 - (y + ofs) * 2];
			bmd->x = 2;
			xprintf("%2d:%s\x13" "A", (y + ofs) * 2, mkval(d * 724 / 1024, 1));
			d = fftout[N_WFM / 2 + (y + ofs) * 2 + 1] + fftout[N_WFM / 2 - (y + ofs) * 2 - 1];
			bmd->x = 70;
			xprintf("%2d:%s\x13" "A", (y + ofs) * 2 + 1, mkval(d * 724 / 1024, 1));
		}
	}
	bmd->palette = Pal2bw;
	disp_transfer(bmd, 0, 0, 127, 95, 0, 32);
}




/*--------------------------------------------------------------------/
/ SCT0 ISR (on falling edge of SCT0_OUT3)                             /
/--------------------------------------------------------------------*/

void SCT0_IRQHandler (void)
{
	static uint8_t soc;
	static uint16_t kdiv, tdiv, pc_cnt, wi, cyc_alt, cyc_time;
	static uint32_t tp_inp, tp_div;
	static int64_t p_acc, c_acc;
	uint32_t freq;
	int32_t volt, curr, d1, d2;
	uint8_t cap, d, evf;


	SCT0_EVFLAG = SCT0_EVFLAG;	/* Clear SCT0 irq */

	Timer++;
	evf = Flags;
	if (!(evf & FLAG_RUN)) return;


	SPI0_TXDATCTL = 0x0F1FFFFF;			/* Start to read data from AD7685 via SPI0 */

	volt = ((ADC0_DAT1 >> 4 & 0x0FFF) ^ 0x0FFF);	/* Get voltage ADC value (12bit) */
	volt = (volt - Cal[0]) * 1024 / Cal[1];			/* Apply offset/gain compensation */

	while (!(SPI0_STAT & _BV(0))) ;		/* Wait for end of SPI transfer */
	curr = (SPI0_RXDAT & 0xFFFF) ^ 0xFFFF;		/* Get current ADC value (16bit) */
	if (curr <= 480 || curr >= 65536 - 260) evf |= FLAG_OVF;	/* Input over range? */
	curr = (curr - Cal[2]) * 1024 / Cal[3];		/* Apply offset/gain compensation */

	/* Measure cycle time for frequency display and resampling */
	cyc_time++;
	if (volt < -100) soc |= 1;
	if ((soc & 1) && Adc_V < 0 && volt >= 0) {	/* Zero-cross detection (up slope) */
		AccJoule += p_acc;
		if (NumCC < 8) {
			NumCC++;
			d = WpCC;
			CycleCurrent[d] = c_acc / (int)pc_cnt;
			WpCC = ++d % 8;
		}
		c_acc = p_acc = pc_cnt = 0;
		evf |= FLAG_CYCLE;	/* AccJoule and CycleCurrent are updated */
		if (++cyc_alt >= N_ALT) {	/* 64 alt deteceted? */
			freq = (F_SAMPLE * N_ALT * 200UL / cyc_time + 1) / 2;
			if (freq < 4500 || freq > 6500) {	/* Freqency out of range? (not in 45-65Hz) */
				FreqRT = 0;		/* Invalidate frequency */
				tp_inp = 0;		/* Suspend re-sampling */
			} else {
				FreqRT = freq;		/* Measured frequency */
				tp_inp = cyc_time;	/* Resampling factor */
			}
			cyc_alt = cyc_time = 0;
			evf |= FLAG_FREQ;	/* FreqRT is updated */
		}
		soc = 0x80;	/* Start of a cycle */
	} else {
		soc &= ~0x80;
	}
	p_acc += volt * curr;
	c_acc += curr * curr;
	pc_cnt++;

	if (tp_inp) {	/* Frequency valid? */
		cap = Capture;
		if (soc & cap & 0x80) {	/* New capture requested? */
			cap &= ~0x80; wi = 0;
		}
		tp_div += N_WFM * N_ALT;	/* Capture a cycle of wave form in resample of N_WFM samples */
		if (tp_div >= tp_inp) {
			tp_div -= tp_inp;
			if (cap > 0 && cap < 0x80) {
				d1 = tp_div; d2 = N_WFM * N_ALT - tp_div;
				Waveform[wi][0] += (Adc_V * d1 + volt * d2) / (N_WFM * N_ALT);
				Waveform[wi][1] += (Adc_I * d1 + curr * d2) / (N_WFM * N_ALT);
				if (++wi >= N_WFM) {
					wi = 0; cap--;
				}
			}
		}
		Capture = cap;
	}

	Adc_V = volt; Adc_I = curr;
	evf |= FLAG_SAMPLE;		/* Adc_V, Adc_I are updated */

	if (++kdiv >= F_SAMPLE / 100) {	/* Key scan process */
		kdiv = 0;
		d = ~GPIO_PIN0HL;
		BtnDown |= (BtnStat ^ d) & d;
		BtnStat = d;
	}

	if (++tdiv >= F_SAMPLE) {		/* Duration timer */
		tdiv = 0;
		TimeRT++;
	}

	Flags = evf;
}



/*--------------------------------------------------------------------/
/ Program Main                                                        /
/--------------------------------------------------------------------*/

int main (void)
{
	uint8_t btn, md, hold = 0;
	int ofs, mag = 6;


	/* Initialize integrated peripherals */
	ioinit();

	/* Initialize physical display */
	BmdLcd.bitmap = 0;
	disp_init(&BmdLcd, 128, 128, 16);	/* Create physical display descriptor */
	xdev_out(putch_bmd);

	load_calib();

	/* Initialize on-memory screen */
	PutchBmd = &BmdMem;
	BmdMem.bitmap = Bitmap;

	capture_wfm(3);

	for (;;) {

		/* Basic load meter */
		rfsh_full(1);
		rfsh_full(0);
		for (;;) {
			if (!hold) {
				capture_wfm(3);
				rfsh_full(0);
			}
			btn = get_btn();
			if (btn == BTN_MODE) break;
			if (btn == BTN_HOLD) hold ^= 1;
		}

		/* Waveform/Harmonic monitor */
		for (md = 0; md < 2; md++) {
			rfsh_simple();
			ofs = 0;
			if (md) {
				rfsh_harmonics(ofs);
			} else {
				rfsh_scope();
			}
			for (;;) {
				if (!hold) {
					capture_wfm(3);
					rfsh_simple();
					if (md) {
						rfsh_harmonics(ofs);
					} else {
						rfsh_scope();
					}
				}
				btn = get_btn();
				if (btn == BTN_MODE) break;
				if (btn == BTN_HOLD) hold ^= 1;
				if (md) {
					if (btn == BTN_UP && ofs) rfsh_harmonics(--ofs);
					if (btn == BTN_DOWN && ofs < 11) rfsh_harmonics(++ofs);
				}
			}
		}

		/* Current profile monitor */
		rfsh_simple();
		rfsh_roll(hold ? 11 : 15, mag);
		for (;;) {
			if (!hold) {
				rfsh_roll(3, mag);
				if (!Capture) {
					capture_wfm(2);
					rfsh_simple();
					capture_wfm(1);
				}
			}
			btn = get_btn();
			if (btn == BTN_MODE) break;
			if (btn == BTN_HOLD) {
				hold ^= 1;
				if (!hold) {
					rfsh_roll(12, mag);
					capture_wfm(1);
				} else {
					capture_wfm(2);
					rfsh_simple();
				}
			}
			if (btn == BTN_UP && mag) {
				rfsh_roll(8, --mag);
				if (hold) rfsh_roll(1, mag);
			}
			if (btn == BTN_DOWN && mag < 7) {
				rfsh_roll(8, ++mag);
				if (hold) rfsh_roll(1, mag);
			}
		}

	}
}

