/*--------------------------------------------------------------/
/  Startup Module for NXP LPC15xx Microcontrollers              /
/                                                               /
/ * This module defines vector table, startup code, default     /
/   exception handlers, main stack and miscellanous functions.  /
/ * This file is a non-copyrighted public domain software.      /
/--------------------------------------------------------------*/

#include "LPC15xx.h"

/*-------------------------------------------------------------------*/
/* Basic Configurations                                              */

#define STACK_SIZE	0	/* Stack size (Multiple of 8. If zero, MSP is set to last RAM address + 1) */

#define	SYSOSC_MODE	1			/* System oscillator mode = 0:Disabled, 1:Xtal, 2:External */
#define	MAIN_SRC	3			/* Main clock source = 0:IRC, 1:System osc, 2:WDT osc or 3:PLL */
#define PLL_SRC		1			/* PLL input source = 0:IRC, 1:System osc */

#define	F_OSC		12000000	/* Oscillator frequency to be used for main clock or PLL-in */
#define	F_SYSCLK	18000000	/* System clock frequency (<= 72M) */
#define	F_MAINCLK	36000000	/* Main clock frequency (multiple of F_OSC and F_SYSCLK, <= 100M) */



/*-------------------------------------------------------------------*/
/* System clock configurations                                       */


#if (F_MAINCLK % F_OSC) || (F_MAINCLK % F_SYSCLK)
#error Invalid combination of F_MAINCLK, F_OSC and F_SYSCLK
#endif
#if !SYSOSC_MODE && (MAIN_SRC == 1 || PLL_SRC == 1)
#error Wrong SYSOSC_MODE setting
#endif
#if SYSOSC_MODE && (F_OSC < 1000000 || F_OSC > 25000000)
#error F_OSC is out of range for system oscillator
#endif
#if MAIN_SRC == 3	/* Use PLL */
#if F_OSC < 10000000
#error F_OSC is too low for PLL input
#endif
#if F_MAINCLK * 2 >= 156000000
#define P_SEL 0
#elif F_MAINCLK * 4 >= 156000000
#define P_SEL 1
#elif F_MAINCLK * 8 >= 156000000
#define P_SEL 2
#else
#define P_SEL 3
#endif
#else
#define P_SEL 0
#endif

static
void configure_sysclk (void)
{
	int n;


	if (SYSOSC_MODE) {	/* Enable system oscillator if needed */
		SYSOSCCTRL = (F_OSC >= 17500000 ? 2 : 0) | (SYSOSC_MODE - 1);	/* Select frequency range and input mode */
		__enable_power(PDSYSOSC);				/* Power-up sysosc */
		for (n = 0; n < 1000; n++) PDRUNCFG;	/* A delay for sysosc stable */
	}

	if (MAIN_SRC == 2)			/* Enable WDT oscillator if needed */
		PDRUNCFG &= ~_BV(20);	/* Power-up WDT osc */

	if (MAIN_SRC == 3) {		/* Configure PLL if needed */
		SYSPLLCLKSEL = PLL_SRC;						/* Select PLL-in source */
		SYSPLLCTRL = (F_MAINCLK / F_OSC - 1) | (P_SEL << 6);	/* Set PLL multiplier */
		__enable_power(PDSYSPLL);					/* Power-up SYSPLL osc */
		while (!(SYSPLLSTAT & _BV(0))) ;			/* Wait for SYSPLL locked */
	}

	SYSAHBCLKDIV = F_MAINCLK / F_SYSCLK;	/* Set system clock divider */

	if (MAIN_SRC == 3)		/* Select main clock source */
		MAINCLKSELB = 2;
	else
		MAINCLKSELA = MAIN_SRC;

	if (F_SYSCLK <= 55)		/* Select flash access time */
		FLASHCFG = (FLASHCFG & ~(3 << 12)) | (F_SYSCLK <= 25 ? 0 : 1) << 12;
}



/*--------------------------------------------------------------------/
/ Declareations                                                       /
/--------------------------------------------------------------------*/

/* Section address defined in linker script */
extern long _sidata[], _sdata[], _edata[], _sbss[], _ebss[], _endof_sram[];
extern int main (void);

/* Exception/IRQ Handlers */
void Reset_Handler (void) __attribute__ ((noreturn));
void NMI_Handler (void) __attribute__ ((weak, alias ("Exception_Trap")));
void HardFault_Hander (void) __attribute__ ((weak, alias ("Exception_Trap")));
void MemManage_Handler (void) __attribute__ ((weak, alias ("Exception_Trap")));
void BusFault_Handler (void) __attribute__ ((weak, alias ("Exception_Trap")));
void UsageFault_Handler (void) __attribute__ ((weak, alias ("Exception_Trap")));
void SVC_Handler (void) __attribute__ ((weak, alias ("Exception_Trap")));
void DebugMon_Handler (void) __attribute__ ((weak, alias ("Exception_Trap")));
void PendSV_Handler (void) __attribute__ ((weak, alias ("Exception_Trap")));
void SysTick_Handler (void) __attribute__ ((weak, alias ("Exception_Trap")));
void WDT_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void BOD_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void FLASH_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void EE_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void DMA_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void GINT0_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void GINT1_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void PIN_INT0_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void PIN_INT1_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void PIN_INT2_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void PIN_INT3_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void PIN_INT4_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void PIN_INT5_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void PIN_INT6_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void PIN_INT7_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void RIT_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void SCT0_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void SCT1_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void SCT2_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void SCT3_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void MRT_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void UART0_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void UART1_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void UART2_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void I2C0_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void SPI0_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void SPI1_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void C_CAN0_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void USB_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void USB_FIQ_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void USB_WAKEUP_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void ADC0_SEQA_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void ADC0_SEQB_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void ADC0_THCMP_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void ADC0_OVR_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void ADC1_SEQA_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void ADC1_SEQB_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void ADC1_THCMP_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void ADC1_OVR_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void DAC_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void CMP0_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void CMP1_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void CMP2_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void CMP3_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void QEI_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void RTC_ALARM_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));
void RTC_WAKE_IRQHandler (void) __attribute__ ((weak, alias ("IRQ_Trap")));



/*--------------------------------------------------------------------/
/ Main Stack                                                          /
/--------------------------------------------------------------------*/

#if STACK_SIZE > 0
static char mstk[STACK_SIZE] __attribute__ ((aligned(8), section(".STACK")));
#define INITIAL_MSP	&mstk[STACK_SIZE]
#else
#define INITIAL_MSP	_endof_sram
#endif



/*--------------------------------------------------------------------/
/ Exception/Interrupt Vector Table                                    /
/--------------------------------------------------------------------*/

/* Vector table to be located at address 0 */
const void* const vector[] __attribute__ ((section(".VECTOR"))) =
{
	INITIAL_MSP, Reset_Handler, NMI_Handler, HardFault_Hander,
	MemManage_Handler, BusFault_Handler, UsageFault_Handler, 0,
	0, 0, 0, SVC_Handler,
	DebugMon_Handler, 0, PendSV_Handler, SysTick_Handler,
	WDT_IRQHandler, BOD_IRQHandler, FLASH_IRQHandler, EE_IRQHandler,
	DMA_IRQHandler, GINT0_IRQHandler, GINT1_IRQHandler, PIN_INT0_IRQHandler,
	PIN_INT1_IRQHandler, PIN_INT2_IRQHandler, PIN_INT3_IRQHandler, PIN_INT4_IRQHandler,
	PIN_INT5_IRQHandler, PIN_INT6_IRQHandler, PIN_INT7_IRQHandler, RIT_IRQHandler,
	SCT0_IRQHandler, SCT1_IRQHandler, SCT2_IRQHandler, SCT3_IRQHandler,
	MRT_IRQHandler, UART0_IRQHandler, UART1_IRQHandler, UART2_IRQHandler,
	I2C0_IRQHandler, SPI0_IRQHandler, SPI1_IRQHandler, C_CAN0_IRQHandler,
	USB_IRQHandler, USB_FIQ_IRQHandler, USB_WAKEUP_IRQHandler, ADC0_SEQA_IRQHandler,
	ADC0_SEQB_IRQHandler, ADC0_THCMP_IRQHandler, ADC0_OVR_IRQHandler, ADC1_SEQA_IRQHandler,
	ADC1_SEQB_IRQHandler, ADC1_THCMP_IRQHandler, ADC1_OVR_IRQHandler, DAC_IRQHandler,
	CMP0_IRQHandler, CMP1_IRQHandler, CMP2_IRQHandler, CMP3_IRQHandler,
	QEI_IRQHandler, RTC_ALARM_IRQHandler, RTC_WAKE_IRQHandler
};



/*---------------------------------------------------------------------/
/ Reset_Handler is the C runtime start-up routine. It configures       /
/ processor core, system clock generator then initialize .data and     /
/ .bss sections and then start main().                                 /
/---------------------------------------------------------------------*/

void Reset_Handler (void)
{
	long *s, *d;

	/* Configure BOD control (Reset/Interrupt at 2.7V/2.9V) */
	BODCTRL = 0x1F;

	configure_sysclk();	/* Configure system clock generator */

	/* Initialize .data/.bss section and static objects get ready to use after this process */
	for (s = _sidata, d = _sdata; d < _edata; *d++ = *s++) ;
	for (d = _sbss; d < _ebss; *d++ = 0) ;

	/* Start main() with MSP */
	main();

	for (;;) ;
}



/*--------------------------------------------------------------------/
/ Unexpected Exception/IRQ Trap                                       /
/--------------------------------------------------------------------*/

void Exception_Trap (void)
{
	for (;;) ;
}


void IRQ_Trap (void)
{
	for (;;) ;
}

